# import libraries
import smtplib
import os
import re
import pathlib
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication
from email.header import Header
import emli

# function to send email from one mail to another mail
def sendMail(senderEmail, recipientEmail, smtpServer, username, appPassword, subject, title, name, bodyHTML,
             filenames, port):

    # generate greeting for email
    greeting = ""
    # consider all possible situations (male and female)
    if title == emli.manTitle:
        greeting = emli.manGreeting + " " + name + ","
    elif title == emli.womanTitle:
        greeting = emli.womanGreeting + " " + name + ","

    # create full body for letter (HTML version)
    fullBodyHTML = ""
    splitBodyHTML = bodyHTML.split("\n")
    # go through all lines of HTML body
    for line in splitBodyHTML:
        # find out row for beginning of HTML body
        if emli.begHTMLBody in line:
            # add additional greeting for HTML body
            fullBodyHTML = fullBodyHTML + line + "\n"
            fullBodyHTML = fullBodyHTML + "        " + greeting + emli.breakSymbol + "\n"
        else:
            # general case
            fullBodyHTML = fullBodyHTML + line + "\n"

    # debug code
    # print(bodyHTML)
    # print(fullBodyHTML)

    # generate message for mail client
    msg = MIMEMultipart()
    msg['From'] = Header(senderEmail)
    msg['To'] = Header(recipientEmail)
    msg['Subject'] = Header(subject)

    # attach two types of content (plain and HTML) for mail client
    # msg.attach(MIMEText(fullBodyPlain, 'plain', 'utf-8'))
    msg.attach(MIMEText(fullBodyHTML, 'html', 'utf-8'))

    # add attachments in letter
    for pathInput in filenames:
        # handle filename for message
        attName = os.path.basename(pathInput)
        # open file
        fileInput = open(pathInput, 'rb')
        # receive extension of file
        ext = pathlib.Path(pathInput).suffix
        ext = ext[1:]
        # attach file in message
        att = MIMEApplication(fileInput.read(), _subtype=ext)
        # close file
        fileInput.close()
        # attach header in message
        att.add_header('Content-Disposition', 'attachment', filename=attName)
        # do final attachment
        msg.attach(att)

    # create a SMTP connection
    server = smtplib.SMTP(smtpServer, port)
    # configure SMTP connection
    server.ehlo()
    server.starttls()
    server.ehlo()

    # log in to the server
    server.login(username, appPassword)
    # define the recipient's email address
    server.sendmail(senderEmail, recipientEmail, msg.as_string())
    # close the connection
    server.quit()



# import library
from urllib.parse import urlparse
import re
import os

### Constant section

## define numeric constants
smtpPort = 587 # port for SMTP protocol

## define SMTP configuration constants
senderEmail = "tomilo.art.forscience@gmail.com"
smtpServer = "smtp.gmail.com"
username = "tomilo.art.forscience@gmail.com"
appPassword = "epbk ardc hgky mnwa"

## define visual interface string constants
filenameBuilder = "emli.glade" # filename for builder file of GTK GUI
nameMainWindow = "mainWindow" # name of main window in graphical user interface
nameMailButton = "mailButton" # name of button in graphical user interface
nameConfFileChooser = "confFileChooser" # name of chooser for configuration file
nameDataFileChooser = "dataFileChooser" # name of chooser for data file
nameContTextField = "contTextField" # name of special text field for
nameAttachFileChooserWidget = "attachFileChooserWidget" # name of chooser widget for attachments of letter
nameSubjectEntry = "subjectEntry" # name of entry field for letter subject
nameErrorLabel = "errorLabel" # name of error label in GUI
nameToolbarTextMod = "toolbarTextMod" # name of toolbar object for text modification
nameBoldText = "bold" # name of bold text in GTK library
nameItalicText = "italic" # name of italic text in GTK library
nameUnderlineText = "underline" # name of underline text in GTK library

## define email content string constants

manTitle = "Herr" # title for men in greeting
womanTitle = "Frau" # title for women in greeting
manGreeting = "Sehr geehrter" # greeting for men in letter
womanGreeting = "Sehr geehrte" # greeting for women in letter
begHTMLBody = "<p>" # indicator row for beginning of HTML body
breakSymbol = "<br>" # indicator of break symbol in HTML file
delimeterTypeContent = "************" # split the content into two parts (plain and HTML)

## define initial and final letter texts for sending mails

initialLetterText = """<html>
  <body>
    <p>"""
finalLetterText = """    </p>
  </body>
</html>"""


## define general string constants

emptyRow = "" # constant for empty row

# define constants for regular expression
regexMail = r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,7}\b'
regexDomain = "[a-zA-Z0-9]+(\.[a-zA-Z]{2,})+$"

## Function section

# function to verify whether string variable contains correct email address
def isValidEmail(addr):
    # check correctness of email address
    if (re.fullmatch(regexMail, addr)):
        return True
    else:
        return False

# function to verify whether string variable contains correct domain address
def isValidDomain(addr):
    # check correctness of domain address
    if (re.match(regexDomain, addr)):
        return True
    else:
        return False

# function to verify correctness of path in file system
def isValidFile(path):
    # check correctness of path
    if os.path.isfile(path):
        return True
    else:
        return False

# function to add headers of HTML text for SMTP protocol
def addHeadersHTML(textHTML):
    # add special headers in text
    modTextHTML = initialLetterText + "\n" + textHTML + "\n" + finalLetterText
    # return modified HTML text
    return modTextHTML



"""
This is EmLi package for sending mails for several users.
The given file is default for Python project and has some basic
documentation.
"""
from .emli import main
from .alltypes import *
from .gui import *
from .reader import *
from .mail import *



# import libraries
import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Pango, GLib
import emli

# variable for text entry objects
subjectEnrty = None
# variable for file chooser objects
dataFileChooser = None
# variable for file chooser widget object
attachFileChooserWidget = None
# variable for text field object
contTextField = None
# variable for label objects
errorLabel = None
# variable for toolbar object (text modification)
toolbarTextMod = None

# function to handle signal of mail sending
def handleSendMail(self):

    # initialization
    errorLabel.set_text(emli.emptyRow)
    # read configuration constants from special module
    senderEmail = emli.senderEmail
    smtpServer = emli.smtpServer
    username = emli.username
    appPassword = emli.appPassword

    # read filenames of different files
    filenameTable = dataFileChooser.get_filename()
    filenamesAttach = attachFileChooserWidget.get_filenames()

    # read HTML body for letter from text field
    textBuffer = contTextField.get_buffer()
    startIter = textBuffer.get_start_iter()
    endIter = textBuffer.get_end_iter()
    formatBuf = textBuffer.register_serialize_tagset()
    exported = textBuffer.serialize(textBuffer, formatBuf, startIter, endIter)
    pangoConverter = emli.PangoToHtml()
    bodyHTML = pangoConverter.feed(exported)
    # modify body HTML text
    bodyHTML = emli.addHeadersHTML(bodyHTML)
    # read subject from special text entry object
    subject = subjectEnrty.get_text()

    # debug code
    # print(bodyHTML)
    # print(subject)

    # read data from table file (CSV)
    try:
        titles, names, recipientEmails = emli.readTableFromCSVFile(filenameTable)
    except Exception as error:
        # handle exception situation
        errorLabel.set_text(str(error))
        # finish working
        return

    # go through all mail boxes
    index = 0
    numMails = len(recipientEmails)
    while index < numMails:
        # extract data from arrays
        recipientEmail = recipientEmails[index]
        name = names[index]
        title = titles[index]
        try:
            # send one mail to end user seperatly
            emli.sendMail(senderEmail, recipientEmail, smtpServer, username, appPassword, subject, title, name, bodyHTML, filenamesAttach, alltypes.smtpPort)
        except Exception as error:
            # handle exception situation
            errorLabel.set_text(str(error))
            # finish working
            return
        # go to the next mail box
        index = index + 1

# function to handle signal of destruction
def handleDestroyButton(self):
    Gtk.main_quit()

# function to handle signal for bold button
def handleBoldButton(self, tag):
    applyTagToTextView(tag)

# function to handle signal for italic button
def handleItalicButton(self, tag):
    applyTagToTextView(tag)

# function to handle signal for underline button
def handleUnderlineButton(self, tag):
    applyTagToTextView(tag)

# function to apply tag to text view object
def applyTagToTextView(tag):

    # get the text buffer from view object
    textBuffer = contTextField.get_buffer()
    # define boundaries of text
    bounds = textBuffer.get_selection_bounds()
    if len(bounds) != 0:
        # apply new tag for text
        start, end = bounds
        textBuffer.apply_tag(tag, start, end)


# function to create builder for GUI app
def createBuilder(filenameBuilder):
    # create builder
    builder = Gtk.Builder()
    builder.add_from_file(filenameBuilder)
    # return builder
    return builder

# function to create tags for text modification
def createTextTags(textField):

    # receive text buffer from text field
    textBuffer = textField.get_buffer()
    # create tags using found buffer
    boldTag = textBuffer.create_tag("bold", weight=Pango.Weight.BOLD)
    italicTag = textBuffer.create_tag("italic", style=Pango.Style.ITALIC)
    underlineTag = textBuffer.create_tag("underline", underline=Pango.Underline.SINGLE)
    # return tags
    return boldTag, italicTag, underlineTag


# function to receive GUI object from builder
def receiveObject(builder, nameObject):
    # find out object
    visObject = builder.get_object(nameObject)
    # return object
    return visObject

# function to create new buttons for toolbar of modification of text
def addNewButtonsToToolbar(toolbarObject):

    # create three buttons (different text styles)
    boldButton = Gtk.ToolButton(Gtk.STOCK_BOLD)
    italicButton = Gtk.ToolButton(Gtk.STOCK_ITALIC)
    underlineButton = Gtk.ToolButton(Gtk.STOCK_UNDERLINE)
    # insert buttons into toolbar object
    toolbarObject.insert(boldButton, 0)
    toolbarObject.insert(italicButton, 1)
    toolbarObject.insert(underlineButton, 2)
    # return create buttons
    return boldButton, italicButton, underlineButton

# function to attach function to click event of visual object
def attachClickEvent(visObject, handlerFunc):
    # attach function to visual object
    visObject.connect("clicked", handlerFunc)

# function to attach function to click event of visual object with one argument
def attachClickEventArg(visObject, handlerFunc, argument):
    # attach function to visual object
    visObject.connect("clicked", handlerFunc, argument)

# function to attach function to destroy event of visual object
def attachDestroyEvent(visObject, handlerFunc):
    # attach function to visual object
    visObject.connect("destroy", handlerFunc)

# function to save references to file chooser objects
def saveRefToFileChooserObjects(ref1):
    # specify global variables
    global dataFileChooser
    # save references in special variables
    dataFileChooser = ref1

# function to save references to file chooser widget objects
def saveRefToFileChooserWidgetObjects(ref1):
    # specify global variables
    global attachFileChooserWidget
    # save references in special variables
    attachFileChooserWidget = ref1

# function to save references to text entry objects
def saveRefToTextEntryObjects(ref1):
    # specify global variables
    global subjectEnrty
    # save references in special variables
    subjectEnrty = ref1

# function to save references to text field objects
def saveRefToTexfFieldObjects(ref1):
    # specify global variables
    global contTextField
    # save references in special variables
    contTextField = ref1

# function to save references to label objects
def saveRefToLabelObjects(ref1):
    # specify global variables
    global errorLabel
    # save references in special variables
    errorLabel = ref1

# function to save references to label objects
def saveRefToToolbarObjects(ref1):
    # specify global variables
    global toolbarTextMod
    # save references in special variables
    toolbarTextMod = ref1

# function to initialize properties of visual objects
def initializeVisualObjects():
    # empty row for error label
    errorLabel.set_text(emli.emptyRow)
    # "convenient" text for letter content
    textBuffer = contTextField.get_buffer()
    textBuffer.set_text("")
    contTextField.set_wrap_mode(Gtk.WrapMode.WORD)

# function to launch graphical user interface (GTK+)
def launchInterface(window):
    # launch interface
    window.show_all()
    Gtk.main()



# import libraries
import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Pango, GLib
import re
import csv
from html.parser import HTMLParser
from typing import Dict, List, Optional, Tuple
from xml.etree.ElementTree import fromstring
import emli

# function to read configuration file from file
def readConfigurationFile(filenameConf):

    # find out number of lines in file
    fileConf = open(filenameConf, "r")
    numLines = len(fileConf.readlines())
    fileConf.close()
    # open file for reading
    fileConf = open(filenameConf, "r")

    # verify number of lines in file
    if numLines < 4:
        # throw exceptional situation
        raise IndexError("File is too small for app configuration!")

    # read configuration data from file
    senderMail = re.sub("\n", "", fileConf.readline())
    smtpServer = re.sub("\n", "", fileConf.readline())
    username = re.sub("\n", "", fileConf.readline())
    appPassword = re.sub("\n", "", fileConf.readline())

    # check correctness of email address
    if not alltypes.isValidEmail(senderMail):
        # throw exceptional situation
        raise ValueError("Sender mail in file is incorrect!")
    # check correctness of domain address
    if not alltypes.isValidDomain(smtpServer):
        # throw exceptional situation
        raise ValueError("Sender mail in file is incorrect!")

    # close file
    fileConf.close()
    # return configuration data
    return senderMail, smtpServer, username, appPassword

# function to read emails from file
def readEmailsFromTextFile(filenameEmail):

    # initialization
    emails = []
    # open file
    fileEmail = open(filenameEmail, "r")

    # go through all lines of file
    for line in fileEmail.readlines():
        # delete extra symbol in line
        line = re.sub("\n", "", line)
        # check correctness of email address
        if alltypes.isValidEmail(line):
            emails.append(line)
        else:
            # throw exceptional situation
            raise ValueError("Recepeint email in file is incorrect!")

    # close file
    fileEmail.close()
    # return array of emails
    return emails

# function to read table from CSV file
def readTableFromCSVFile(filenameTable):

    # initialization
    titles = []
    names = []
    emails = []
    index = 1
    # open file
    fileTable = open(filenameTable, "r")

    # create reader using file object
    reader = csv.reader(fileTable)
    # go through all rows of table
    for row in reader:
        # the first row will be ignored
        if index != 1:
            # add data in special arrays
            titles.append(row[0])
            names.append(row[1] + " " + row[2])
            emails.append(row[3])
        # add index to variable
        index = index + 1

    # return array of emails
    return titles, names, emails

# function to read subject and body from file
def readSubjectAndBodyFromFile(filenameContent):

    # find out number of lines in file
    fileCont = open(filenameContent, "r")
    numLines = len(fileCont.readlines())
    fileCont.close()

    # verify number of lines in file
    if numLines < 2:
        # throw exceptional situation
        raise IndexError("File is too small for email content!")

    # initialization
    subject = ""
    bodyPlain = ""
    i = 1
    ind = False
    # open file (plain version)
    fileCont = open(filenameContent, "r")

    # go through all lines of file
    for line in fileCont.readlines():
        # delete extra symbol in line
        line = re.sub("\n", "", line)
        # consider different situations
        if i == 1:
            # subject situation
            subject = line
        else:
            # find out delimeter row
            if line == alltypes.delimeterTypeContent:
                ind = True
            # body situation (plain text)
            if ind == False:
                bodyPlain = bodyPlain + line + "\n"
        i = i + 1


    # close file
    fileCont.close()
    # open file again
    fileCont = open(filenameContent, "r")
    # initialization (HTML version)
    ind = False
    bodyHTML = ""

    # go through all lines of file
    for line in fileCont.readlines():
        # delete extra symbol in line
        line = re.sub("\n", "", line)
        # find out delimeter row
        if line == alltypes.delimeterTypeContent:
            # change value of indicator
            ind = True
            # go to the next row
            continue
        # body situation (HTML text)
        if ind == True:
            bodyHTML = bodyHTML + line + "\n"

    # close file
    fileCont.close()

    # return subject and body
    return subject, bodyPlain, bodyHTML


# function to read subject and body from file
def readFilenamesFromFile(filename):

    # initialization
    filenames = []
    # open file
    fileInput = open(filename, "r")

    # go through all lines of file
    for line in fileInput.readlines():
        # delete extra symbol in line
        line = re.sub("\n", "", line)
        # check correctness of file
        if alltypes.isValidFile(line):
            # add line in special array
            filenames.append(line)
        else:
            # throw exceptional situation
            raise ValueError("Filename for attachment is not correct!")

    # close file
    fileInput.close()
    # return filenames that were read
    return filenames

# class to convert Pango to HTML code
class PangoToHtml(HTMLParser):
    """Decode a subset of Pango markup and serialize it as HTML.

    Only the Pango markup used within Gourmet is handled, although expanding it
    is not difficult.

    Due to the way that Pango attributes work, the HTML is not necessarily the
    simplest. For example italic tags may be closed early and reopened if other
    attributes, eg. bold, are inserted mid-way:

        <i> italic text </i><i><u>and underlined</u></i>

    This means that the HTML resulting from the conversion by this object may
    differ from the original that was fed to the caller.
    """
    def __init__(self):
        super().__init__()
        self.markup_text:           str  = ""  # the resulting content
        self.current_opening_tags:  str  = ""  # used during parsing
        self.current_closing_tags:  List = []  # used during parsing

        # the key is the Pango id of a tag, and the value is a tuple of opening
        # and closing html tags for this id.
        self.tags: Dict[str: Tuple[str, str]] = {}

    tag2html: Dict[str, Tuple[str, str]] = {
                                            Pango.Style.ITALIC.value_name:      ("<i>", "</i>"),  # Pango doesn't do <em>
                                            str(Pango.Weight.BOLD.real):        ("<b>", "</b>"),
                                            Pango.Underline.SINGLE.value_name:  ("<u>", "</u>"),
                                            "foreground-gdk":                   (r'<span foreground="{}">', "</span>"),
                                            "background-gdk":                   (r'<span background="{}">', "</span>")
                                            }

    def feed(self, data: bytes) -> str:
        """Convert a buffer (text and and the buffer's iterators to html string.

        Unlike an HTMLParser, the whole string must be passed at once, chunks
        are not supported.
        """
        # Remove the Pango header: it contains a length mark, which we don't
        # care about, but which does not necessarily decodes as valid char.
        header_end  = data.find(b"<text_view_markup>")
        data        = data[header_end:].decode()

        # get the tags
        tags_begin  = data.index("<tags>")
        tags_end    = data.index("</tags>") + len("</tags>")
        tags        = data[tags_begin:tags_end]
        data        = data[tags_end:]

        # get the textual content
        text_begin  = data.index("<text>")
        text_end    = data.index("</text>") + len("</text>")
        text        = data[text_begin:text_end]

        # convert the tags to html.
        # we know that only a subset of HTML is handled in Gourmet:
        # italics, bold, underlined and normal

        root            = fromstring(tags)
        tags_name       = list(root.iter('tag'))
        tags_attributes = list(root.iter('attr'))
        tags            = [ [tag_name, tag_attribute] for tag_name, tag_attribute in zip(tags_name, tags_attributes)]

        tags_list = {}
        for tag in tags:
            opening_tags = ""
            closing_tags = ""

            tag_name    = tag[0].attrib['name']
            vtype       = tag[1].attrib['type']
            value       = tag[1].attrib['value']
            name        = tag[1].attrib['name']

            if vtype == "GdkColor":  # convert colours to html
                if name in ['foreground-gdk', 'background-gdk']:
                    opening, closing = self.tag2html[name]
                    hex_color = f'{value.replace(":","")}' # hex color already handled by gtk.gdk.color.to_string() method
                    opening = opening.format(hex_color)
                else:
                    continue  # no idea!
            else:
                opening, closing = self.tag2html[value]

            opening_tags += opening
            closing_tags = closing + closing_tags   # closing tags are FILO

            tags_list[tag_name] = opening_tags, closing_tags

            if opening_tags:
                tags_list[tag_name] = opening_tags, closing_tags

        self.tags = tags_list

        # create a single output string that will be sequentially appended to
        # during feeding of text. It can then be returned once we've parse all
        self.markup_text                = ""
        self.current_opening_tags       = ""
        self.current_closing_tags       = []  # Closing tags are FILO

        super().feed(text)

        return self.markup_text

    def handle_starttag(self, tag: str, attrs: List[Tuple[str, str]]) -> None:
        # the pango tags are either "apply_tag", or "text"
        # we only really care about the "apply_tag"
        # there could be an assert, but we let the parser quietly handle nonsense.
        if tag == "apply_tag":
            attrs       = dict(attrs)
            tag_name    = attrs.get('name')
            tags        = self.tags.get(tag_name)

            if tags is not None:
                (self.current_opening_tags, closing_tag) = tags
                self.current_closing_tags.append(closing_tag)

    def handle_data(self, data: str) -> None:
        data = self.current_opening_tags + data
        self.markup_text += data

    def handle_endtag(self, tag: str) -> None:
        if self.current_closing_tags:  # Can be empty due to closing "text" tag
            self.markup_text += self.current_closing_tags.pop()
        self.current_opening_tags = ""



# configure local environment
import sys
import os
# configure system path
sys.path.append(os.getcwd())

# import local libraries
import emli 

# entry point for EmLi package
def main():
    # debug code
    entryFilePath = os.path.dirname(os.path.abspath(__file__))
    # create builder
    builder = emli.createBuilder(os.path.join(entryFilePath, emli.filenameBuilder))
    # create main GUI objects
    mainWindow = emli.receiveObject(builder, emli.nameMainWindow)
    mailButton = emli.receiveObject(builder, emli.nameMailButton)
    # create file chooser objects
    dataFileChooser = emli.receiveObject(builder, emli.nameDataFileChooser)
    # create text entry field objects
    subjectEnrty = emli.receiveObject(builder, emli.nameSubjectEntry)
    # create text field objects
    contTextField = emli.receiveObject(builder, emli.nameContTextField)
    # create file chooser wigget objects
    attachFileChooserWidget = emli.receiveObject(builder, emli.nameAttachFileChooserWidget)
    # create label objects
    errorLabel = emli.receiveObject(builder, emli.nameErrorLabel)
    # create toolbar for text modification
    toolbarTextMod = emli.receiveObject(builder, emli.nameToolbarTextMod)
    # create new tags for text modification
    boldTag, italicTag, underlineTag = emli.createTextTags(contTextField)
    # add new buttons to toolbar object
    boldButton, italicButton, underlineButton = emli.addNewButtonsToToolbar(toolbarTextMod)
    # add handlers to visual objects
    emli.attachDestroyEvent(mainWindow, emli.handleDestroyButton)
    emli.attachClickEvent(mailButton, emli.handleSendMail)
    emli.attachClickEventArg(boldButton, emli.handleBoldButton, boldTag)
    emli.attachClickEventArg(italicButton, emli.handleItalicButton, italicTag)
    emli.attachClickEventArg(underlineButton, emli.handleUnderlineButton, underlineTag)
    # save references to file chooser objects in GUI module
    emli.saveRefToFileChooserObjects(dataFileChooser)
    # save references to text entry objects in GUI module
    emli.saveRefToTextEntryObjects(subjectEnrty)
    # save references to file chooser widget objects in GUI module
    emli.saveRefToFileChooserWidgetObjects(attachFileChooserWidget)
    # save references to label objects in GUI module
    emli.saveRefToLabelObjects(errorLabel)
    # save references to text field objects in GUI module
    emli.saveRefToTexfFieldObjects(contTextField)
    # save references to toolbar objects in GUI module
    emli.saveRefToToolbarObjects(toolbarTextMod)
    # initialize visual objects for GUI
    emli.initializeVisualObjects()
    # launch graphical interface
    emli.launchInterface(mainWindow)

# specify entry point
if __name__ == "__main__":
    main()



from setuptools import find_packages, setup

# define constants for working
NAME = 'EmLiLetter'
VERSION = '0.0.1'
DESCRIPTION = 'EmLi software to send mails for several users'
AUTHOR = 'Artem Tomilo'
REQUIRED = ['gi', 'pygobject']
URL = 'https://gitlab.com/bridgearchitect/emli'
LICENSE = 'MIT'
PACKAGES = ['emli']
PACKAGE_DATA = 'emli.glade'

# start setupping process for PyPi repository
setup(
   name=NAME,
   version=VERSION,
   description=DESCRIPTION,
   author=AUTHOR,
   url=URL,
   license=LICENSE,
   packages=PACKAGES,
   install_requires=REQUIRED,
   package_data={'emli': [PACKAGE_DATA]},
   entry_points = {
   'console_scripts': ['emli=emli.emli:main'],
   },
)
